/* javagen.c */

#include "opldefs.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
                  
extern int verbose;
extern int generate_debug;
extern int compile_only;
extern int translate_only;
extern int throwback;

FILE *out;

int conditional = 0;
int returned = 0;

extern void putindent( int indent );
extern void putline( const char* line );
extern void puttext( const char* line );
void outputJavaProgram        ( program *prog );
void outputJavaApplet         ( program *prog );
void outputJavaFieldsForMethod( procedure *proc, int indent );
void outputJavaMethod         ( procedure *proc, int indent );
void outputJavaStatementList  ( statementList *pStmts, int indent );
void outputJavaStatement      ( statement *stmt, int indent );
void outputJavaExpression     ( expression *expr );
void outputJavaIdentType      ( ident* id );

const char *getClassname( program *prog );

extern const char *progname;


void outputJavaProgram( program *prog )
{
  procedureList *pProc;

  if( !prog )
    return;

  putline( "import psion.opl.*;" );
  putline( "" );
  fprintf( out, "public class %s\n", getClassname( prog ) );
  putline( "{" );
  putline( "  private static OPL opl;" );
  putline( "" );

  for( pProc = prog->pProcs; pProc; pProc = pProc->tail )
  {
    if( pProc->head )
      outputJavaFieldsForMethod( pProc->head, 1 );
  }

  putline( "" );

  for( pProc = prog->pProcs; pProc; pProc = pProc->tail )
  {
    if( pProc->head )
    {
      outputJavaMethod( pProc->head, 1 );
      putline( "" );
    }
  }

  putline( "  public static void main( String[] args )" );
  putline( "  {" );
  fprintf( out, "    opl = new OPL( \"%s\" );\n", getClassname( prog ) );
  putline( "" );
  if( prog->pProcs && prog->pProcs->head )
  {
    fprintf( out, "      new %s().%s();\n",
		getClassname( prog ), prog->pProcs->head->name->name );
  }
  putline( "  }" );
  putline( "}" );
}


void outputJavaApplet( program *prog )
{
	procedureList *pProc;

	if( !prog )
		return;

	putline( "import psion.opl.*;" );
	putline( "import java.applet.*;" );
	putline( "" );
	fprintf( out, "public class %s\n", getClassname( prog ) );
	putline( "  extends Applet" );
	putline( "  implements Runnable" );
	putline( "{" );
    putline( "  private static OPL opl;" );
    putline( "" );

	for( pProc = prog->pProcs; pProc; pProc = pProc->tail )
	{
		if( pProc->head )
			outputJavaFieldsForMethod( pProc->head, 1 );
	}

	putline( "" );

	for( pProc = prog->pProcs; pProc; pProc = pProc->tail )
	{
		if( pProc->head )
		{
			outputJavaMethod( pProc->head, 1 );
			putline( "" );
		}
	}

	putline( "  private Thread thread;" );
	putline( "" );
	putline( "  public void init()" );
	putline( "  {" );
	putline( "    opl = new OPL( this );" );
	putline( "" );
	putline( "      requestFocus();" );
	putline( "  }" );
	putline( "" );
	putline( "  public void start()" );
	putline( "  {" );
	putline( "    if( thread != null )" );
	putline( "    {" );
	putline( "      thread.stop();" );
	putline( "      thread = null;" );
	putline( "    }" );
	putline( "" );
	putline( "    thread = new Thread( this );" );
	putline( "    thread.start();" );
	putline( "    requestFocus();" );
	putline( "  }" );
	putline( "" );
	putline( "  public void stop()" );
	putline( "  {" );
	putline( "    if( thread != null )" );
	putline( "    {" );
	putline( "      thread.stop();" );
	putline( "      thread = null;" );
	putline( "    }" );
	putline( "  }" );
	putline( "" );
	putline( "  public void destroy()" );
	putline( "  {" );
	putline( "    stop();" );
	putline( "  }" );
	putline( "" );
	putline( "  public void run()" );
	putline( "  {" );
	putline( "    try" );
	putline( "    {" );

	if( prog->pProcs && prog->pProcs->head )
	{
		fprintf( out, "      %s();\n", prog->pProcs->head->name->name );
	}

	putline( "      System.out.println( \"Program exited.\" );" );
	putline( "    }" );
	putline( "    catch( Throwable e )" );
	putline( "    {" );
	putline( "      System.out.println( \"Runtime Error: \" + e );" );
	putline( "      e.printStackTrace();" );
	putline( "      System.out.println( \"Program has exited abnormally.\" );" );
	putline( "    }" );
	putline( "  }" );
	putline( "" );
	putline( "  public void paint( java.awt.Graphics g )" );
	putline( "  {" );
	putline( "    opl.screen.paint( g );" );
	putline( "  }" );
	putline( "}" );
}


void outputJavaFieldsForMethod( procedure *proc, int indent )
{
  statementList *pStmt;
  identList *pVars;
  expressionList *pArray;
  int first = TRUE;

  if( !proc )
    return;

  for( pStmt = proc->pStmts; pStmt; pStmt = pStmt->tail )
  {
    if( pStmt->head )
    {
      switch( pStmt->head->kind )
      {
      case global_stmt:
        for( pVars = pStmt->head->data.global_and_local_part.pVars;
             pVars; pVars = pVars->tail )
        {
          if( pVars->head )
          {
			putindent( indent );
            outputJavaIdentType( pVars->head );
            fprintf( out, " %s", pVars->head->name );
            if( pVars->head->pArrays )
            {
              for( pArray = pVars->head->pArrays;
                   pArray; pArray = pArray->tail )
              {
                if( pArray->head && ( pVars->head->type != string_id ) )
                {
                  fprintf( out, "[" );
                  fprintf( out, "]" );
                }
              }
              fprintf( out, " = new ", pVars->head->name );
              outputJavaIdentType( pVars->head );
			  for( pArray = pVars->head->pArrays;
			  	   pArray; pArray = pArray->tail )
			  {
				if( pArray->head )
				{
				  if( pVars->head->type == string_id )
		  		  {
					fprintf( out, "()" );
				  }
				  else
				  {
					fprintf( out, "[" );
			  		outputJavaExpression( pArray->head );
					puttext( " + 1" ); /* ### */
					fprintf( out, "]" );
				  }
				}
				}
            }
			if( (pVars->head->type == integer_id) && !pVars->head->pArrays )
				fprintf( out, " = 0" );
            fprintf( out, ";\n" );
          }
        }
        break;
      }
    }
  }
}


void outputJavaMethod( procedure *proc, int indent )
{
  identList *pArg;
  int first = TRUE;

  if( !proc )
    return;

  conditional = 0;
  returned = 0;

  putindent( indent );
  outputJavaIdentType( proc->name );
  fprintf( out, " %s(", proc->name->name );

  if( proc->pArgs )
  {
	fprintf( out, " " );
    for( pArg = proc->pArgs; pArg; pArg = pArg->tail )
    {
      if( pArg->head )
      {
        if( !first )
          fprintf( out, ", " );

        outputJavaIdentType( pArg->head );
        fprintf( out, " %s", pArg->head->name );
      }
      first = FALSE;
    }
	fprintf( out, " " );
  }

  fprintf( out, ")\n" );
  putindent( indent );
  putline( "{" );
  outputJavaStatementList( proc->pStmts, indent+1 );

  if( !returned && (proc->name->type != untyped_id) )
  {
	putindent( indent );
    putline( "  return 0;" );
  }

  putindent( indent );
  putline( "}" );
}
 


void outputJavaStatementList( statementList *pStmts, int indent )
{
  statementList *pStmt;

  if( !pStmts )
    return;

  for( pStmt = pStmts; pStmt; pStmt = pStmt->tail )
  {
    if( pStmt->head )
      outputJavaStatement( pStmt->head, indent );
  }
}



void outputJavaStatement( statement *stmt, int indent )
{
  expressionList *pExpr;
  identList *pVars;
  expressionList *pArray;
  int first = TRUE;

  if( !stmt )
    return;

  switch( stmt->kind )
  {
  case if_stmt:
   {
    int ocond = conditional;
	conditional = 1;
    putindent( indent ); puttext( "if( " );
	if( stmt->data.if_part.condition->type != boolean_type )
	  puttext( "( " );
    outputJavaExpression( stmt->data.if_part.condition );
	if( stmt->data.if_part.condition->type != boolean_type )
	  puttext( " ) != 0" );
    putline( " )" );
    if( stmt->data.if_part.pThen )
    {
      if( stmt->data.if_part.pThen->tail
          || ( stmt->data.if_part.pThen->head
               && ( stmt->data.if_part.pThen->head->kind == if_stmt ) ) )
      {
        putindent( indent ); putline( "{" );
        outputJavaStatementList( stmt->data.if_part.pThen, indent+1 );
        putindent( indent ); putline( "}" );
      }
      else
        outputJavaStatementList( stmt->data.if_part.pThen, indent+1 );
    }
    if( stmt->data.if_part.pElse )
    {
      putindent( indent ); putline( "else" );
      if( stmt->data.if_part.pElse->tail
          || ( stmt->data.if_part.pElse->head
               && ( stmt->data.if_part.pElse->head->kind == if_stmt ) ) )
      {
        putindent( indent ); putline( "{" );
        outputJavaStatementList( stmt->data.if_part.pElse, indent+1 );
        putindent( indent ); putline( "}" );
      }
      else
        outputJavaStatementList( stmt->data.if_part.pElse, indent+1 );
    }
    conditional = ocond;
   }
    break;
  case while_stmt:
    putindent( indent ); puttext( "while( " );
	if( stmt->data.while_and_do_part.condition->type != boolean_type )
	  puttext( "( " );
    outputJavaExpression( stmt->data.while_and_do_part.condition );
	if( stmt->data.while_and_do_part.condition->type != boolean_type )
	  puttext( " ) != 0" );
    putline( " )" );
    if( stmt->data.while_and_do_part.pBody )
    {
      if( stmt->data.while_and_do_part.pBody->tail )
      {
        putindent( indent ); putline( "{" );
        outputJavaStatementList( stmt->data.while_and_do_part.pBody, indent+1 );
        putindent( indent ); putline( "}" );
      }
      else
        outputJavaStatementList( stmt->data.while_and_do_part.pBody, indent+1 );
    }
    break;
  case do_stmt:
    putindent( indent ); putline( "do" );
    if( stmt->data.while_and_do_part.pBody )
    {
      if( stmt->data.while_and_do_part.pBody->tail )
      {
        putindent( indent ); putline( "{" );
        outputJavaStatementList( stmt->data.while_and_do_part.pBody, indent+1 );
        putindent( indent ); putline( "}" );
      }
      else
        outputJavaStatementList( stmt->data.while_and_do_part.pBody, indent+1 );
    }
    putindent( indent ); puttext( "while( " );
	if( stmt->data.while_and_do_part.condition->type == boolean_type )
    {
      puttext( "!( " );
      outputJavaExpression( stmt->data.while_and_do_part.condition );
      putline( " ) );" );
    }
    else
    {
	  puttext( "( " );
      outputJavaExpression( stmt->data.while_and_do_part.condition );
	  putline( " ) == 0 );" );
    }
    break;
  case assignment_stmt:
    putindent( indent );
    if( ( stmt->data.assignment_part.name->type == string_id )
        && !stmt->data.assignment_part.name->pArrays )
    {
      fprintf( out, "%s", stmt->data.assignment_part.name->name );
      fprintf( out, " = " );
      outputJavaExpression( stmt->data.assignment_part.value );
    }
    else
    {
      fprintf( out, "%s", stmt->data.assignment_part.name->name );
      if( stmt->data.assignment_part.name->pArrays )
      {
        for( pArray = stmt->data.assignment_part.name->pArrays;
             pArray; pArray = pArray->tail )
        {
          if( pArray->head )
          {
            fprintf( out, "[ " );
            outputJavaExpression( pArray->head );
            fprintf( out, " ]" );
          }
        }
      }
      fprintf( out, " = " );
	  if( stmt->data.assignment_part.name->type != stmt->data.assignment_part.value->type )
      {
        switch( stmt->data.assignment_part.name->type )
        {
	    case string_id:
		    break;
		case integer_id:
		case floating_point_id:
		case long_integer_id:
			if( stmt->data.assignment_part.value->type != untyped_id )
			{
				puttext( "(" );
				outputJavaIdentType( stmt->data.assignment_part.name );
				puttext( ") " );
				printf( "%s: WARNING: Casting from %s to %s\n", progname,
					typeToString( stmt->data.assignment_part.value->type ),
				    typeToString( stmt->data.assignment_part.name->type ) );
				break;
			} /* else fall thru... */
        case untyped_id:
        default:
			puttext( "/*### (cast?)*/ " );
			printf( "%s: WARNING: Cast from %s to %s needed?\n", progname,
					typeToString( stmt->data.assignment_part.value->type ),
				    typeToString( stmt->data.assignment_part.name->type ) );
        }
		puttext( "( " );
        outputJavaExpression( stmt->data.assignment_part.value );
		puttext( " )" );
      }
	  else
        outputJavaExpression( stmt->data.assignment_part.value );
    }
    fprintf( out, ";\n" );
    break;
  case global_stmt:
    break;
  case local_stmt:
    for( pVars = stmt->data.global_and_local_part.pVars;
         pVars; pVars = pVars->tail )
    {
      if( pVars->head )
      {
        putindent( indent );
        outputJavaIdentType( pVars->head );
        fprintf( out, " %s", pVars->head->name );
        if( pVars->head->pArrays )
        {
          for( pArray = pVars->head->pArrays;
               pArray; pArray = pArray->tail )
          {
            if( pArray->head && ( pVars->head->type != string_id ) )
            {
              fprintf( out, "[" );
              fprintf( out, "]" );
            }
          }
          fprintf( out, " = new ", pVars->head->name );
          outputJavaIdentType( pVars->head );
          for( pArray = pVars->head->pArrays; pArray; pArray = pArray->tail )
          {
            if( pArray->head )
			{
			  if( pVars->head->type == string_id )
		  	  {
                fprintf( out, "()" );
			  }
			  else
              {
                fprintf( out, "[" );
                outputJavaExpression( pArray->head );
                fprintf( out, "]" );
              }
            }
          }
        }
		if( (pVars->head->type == integer_id) && !pVars->head->pArrays )
			fprintf( out, " = 0" );
        fprintf( out, ";\n" );
      }
    }
    break;
  case call_stmt:
    putindent( indent );
    if( stmt->data.call_part.opl_call )
      fprintf( out, "opl." );
    fprintf( out, "%s( ", stmt->data.call_part.name->name );
    for( pExpr = stmt->data.call_part.pArgs; pExpr; pExpr = pExpr->tail )
    {
      if( pExpr->head )
      {
        if( !first )
          fprintf( out, ", " );

        outputJavaExpression( pExpr->head );
      }
      first = FALSE;
    }
    fprintf( out, " );\n" );
    break;
  case return_stmt:
    if( !conditional )
      returned = 1;
    putindent( indent );
    fprintf( out, "return " );
    if( stmt->data.return_part.expr )
      outputJavaExpression( stmt->data.return_part.expr );
	else 
	  fprintf( out, "0" );

    fprintf( out, ";\n" );
   break;
  default:
	  printf( "WARNING: Unknown statement of type %d skipped!\n", stmt->kind );
	  break;
  }
}
 


void outputJavaExpression( expression *expr )
{
  expressionList *pExpr;
  expressionList *pArray;
  int first = TRUE;

  if( !expr )
    return;

  switch( expr->kind )
  {
  case variable_expr:
    fprintf( out, "%s", expr->data.variable_part->name );
    if( expr->data.variable_part->pArrays )
    {
		if( expr->data.variable_part->type == string_id )
		{
			fprintf( out, ".charAt( " );
			if( expr->data.variable_part->pArrays->head )
				outputJavaExpression( expr->data.variable_part->pArrays->head );
			fprintf( out, " )" );
		}
		else
		{
			for( pArray = expr->data.variable_part->pArrays;
			   pArray; pArray = pArray->tail )
			{
				if( pArray->head )
				{
					fprintf( out, "[ " );
					outputJavaExpression( pArray->head );
					fprintf( out, " ]" );
				}
			}
		}
    }
    break;
  case integer_expr:
    fprintf( out, "%d", expr->data.integer_part );
    break;
  case hexint_expr:
    fprintf( out, "0x%x", expr->data.integer_part );
    break;
  case character_expr:
    fprintf( out, "'%c'", expr->data.integer_part );
    break;
  case floating_point_expr:
    fprintf( out, "%f", expr->data.floating_point_part );
    break;
  case long_integer_expr:
    fprintf( out, "%d", (int) expr->data.long_integer_part );
    break;
  case string_expr:
    fprintf( out, "\"%s\"", expr->data.string_part );
    break;
  case addition_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " + " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case subtraction_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " - " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case multiplication_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " * " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case division_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " / " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case equality_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " == " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case inequality_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " != " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case lessthan_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " < " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case greaterthan_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " > " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case lessequal_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " <= " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case greaterequal_expr:
    outputJavaExpression( expr->data.binary_op_part.left );
    fprintf( out, " >= " );
    outputJavaExpression( expr->data.binary_op_part.right );
    break;
  case and_expr:
    if( expr->type == integer_type )
    {
      outputJavaExpression( expr->data.binary_op_part.left );
      fprintf( out, " & " );
      outputJavaExpression( expr->data.binary_op_part.right );
    }
    else
    {
      if( expr->data.binary_op_part.left->type != boolean_type )
        puttext( "( " );
      outputJavaExpression( expr->data.binary_op_part.left );
      if( expr->data.binary_op_part.left->type != boolean_type )
        puttext( " ) != 0" );
      puttext( " && " );
      if( expr->data.binary_op_part.right->type != boolean_type )
        puttext( "( " );
      outputJavaExpression( expr->data.binary_op_part.right );
      if( expr->data.binary_op_part.right->type != boolean_type )
        puttext( " ) != 0" );
    }
    break;
  case or_expr:
    if( expr->type == integer_type )
    {
      outputJavaExpression( expr->data.binary_op_part.left );
      fprintf( out, " | " );
      outputJavaExpression( expr->data.binary_op_part.right );
    }
    else
    {
      if( expr->data.binary_op_part.left->type != boolean_type )
        puttext( "( " );
      outputJavaExpression( expr->data.binary_op_part.left );
      if( expr->data.binary_op_part.left->type != boolean_type )
        puttext( " ) != 0" );
      puttext( " || " );
      if( expr->data.binary_op_part.right->type != boolean_type )
        puttext( "( " );
      outputJavaExpression( expr->data.binary_op_part.right );
      if( expr->data.binary_op_part.right->type != boolean_type )
        puttext( " ) != 0" );
    }
    break;
  case uminus_expr:
    fprintf( out, "- " );
    outputJavaExpression( expr->data.unary_op_part.expr );
    break;
  case parenthesis_expr:
    fprintf( out, "( " );
    outputJavaExpression( expr->data.unary_op_part.expr );
    fprintf( out, " )" );
    break;
  case call_expr:
    if( expr->data.call_part.opl_call )
      fprintf( out, "opl." );
    fprintf( out, "%s( ", expr->data.call_part.name->name );
    for( pExpr = expr->data.call_part.pArgs; pExpr; pExpr = pExpr->tail )
    {
      if( !first )
        fprintf( out, ", " );

      if( pExpr->head )
        outputJavaExpression( pExpr->head );

      first = FALSE;
    }
    fprintf( out, " )" );
    break;
  case power_expr:
	puttext( "Math.pow( " );
    outputJavaExpression( expr->data.binary_op_part.left );
    puttext( ", " );
    outputJavaExpression( expr->data.binary_op_part.right );
	puttext( " )" );
    break;
  default:
	  printf( "WARNING: Unknown expression of type %d skipped!\n", expr->kind );
	  break;
  }
}



void outputJavaIdentType( ident* id )
{
  switch( id->type )
  {
  case untyped_id:
    fprintf( out, "void" );
    break;
  case integer_id:
    fprintf( out, "int" );
    break;
  case long_integer_id:
    fprintf( out, "long" );
    break;
  case floating_point_id:
    fprintf( out, "double" );
    break;
  case string_id:
    /*if( id->pArrays )
      fprintf( out, "char" );
    else*/
      fprintf( out, "String" );
    break;
  }
}


const char *getClassname( program *prog )
{
	char *classname;
	int start, end;

	for( end = strlen( prog->sourcefile ); (end > 0) && (prog->sourcefile[end] != '.'); end-- );
	end--; /* before '.' */
	for( start = end-1; (start >= 0) && (prog->sourcefile[start] != '.') && (prog->sourcefile[start] != '/') && (prog->sourcefile[start] != '\\'); start-- );
	start++; /* after sep */

	classname = (char*) malloc( end - start + 2 );

	strncpy( classname, &prog->sourcefile[start], end - start + 1 );
	classname[end - start + 1] = '\0';

	return classname;
}


void generateJavaCode( program *prog, int applet )
{
	char command[1024];

	if( !prog ) return;

	/* No output filename given, so generate one */
	if( !prog->outputfile )
	{
		char *outfile;
		int n;

		for( n = strlen( prog->sourcefile ); (n > 0) && (prog->sourcefile[n] != '.'); n-- );

		outfile = (char*) malloc( n + 5 );

		strncpy( outfile, prog->sourcefile, n );
		strcpy( &outfile[n], ".java" );

		prog->outputfile = outfile;
	}

	out = fopen( prog->outputfile, "w" );

	if( !out ) return;

	if( applet )
		outputJavaApplet( prog );
	else
		outputJavaProgram( prog );

	fclose( out );

	if( !translate_only )
	{
		sprintf( command, "javac %s\n", prog->outputfile );

		if( verbose )
			fprintf( stderr, command );

		system( command );
	}
}

