/* semantic.c */

#include "opldefs.h"
#include <stdio.h>

void semanticCheckObjectCallStmt( Stmt *stmt, DeclList *pScope );
void semanticCheckDecl( Decl *decl, int *var_offset, DeclList *pScope );
void semanticCheckDeclList( DeclList *pDecl, DeclList *pScope );
void semanticCheckStmt( Stmt *stmt, DeclList *pScope );
void semanticCheckStmtList( StmtList *pStmt, DeclList *pScope );

void semanticCheckObjectCallStmt( Stmt *stmt, DeclList *pScope )
{
  if( pScope && pScope->head )
  {
    DeclList *pDecl = pScope->head->pDecls;
    int found = FALSE;
    for( ; pDecl && !found; pDecl = pDecl->tail )
    {
      if( pDecl->head )
      {
        if( strcmp( pDecl->head->name, stmt->data.object_call_stmt.object ) == 0 )
        {
          found = TRUE;
          stmt->data.object_call_stmt.var = pDecl->head;
        }
      }
    }
  }
}

void semanticCheckDecl( Decl *decl, int *var_offset, DeclList *pScope )
{
  DeclList *temp;

  switch( decl->kind )
  {
  case type_decl:
    break;
  case variable_decl:
    decl->offset = (*var_offset);
    (*var_offset) += 4;
    break;
  case function_decl:
  case procedure_decl:
    break;
  } /* switch */

  temp = (DeclList*) malloc( sizeof( DeclList ) );
  temp->head = decl;
  temp->tail = pScope;
  pScope = temp;

  semanticCheckStmtList( decl->pStmts, pScope );

  temp = pScope;
  pScope = pScope->tail;
  free( temp );
}

void semanticCheckDeclList( DeclList *pDecl, DeclList *pScope )
{
  int var_offset = 0;

  /* calculate offsets for global variables */
  for(; pDecl; pDecl = pDecl->tail )
  {
    if( pDecl->head )
      semanticCheckDecl( pDecl->head, &var_offset, pScope );
  } /* for */
}

void semanticCheckStmt( Stmt *stmt, DeclList *pScope )
{
  switch( stmt->kind )
  {
  case global_call_stmt:
  case self_call_stmt:
    break;
  case object_call_stmt:
    semanticCheckObjectCallStmt( stmt, pScope );
    break;
  default:
    break;
  }
}

void semanticCheckStmtList( StmtList *pStmt, DeclList *pScope )
{
  /* calculate offsets for global variables */
  for(; pStmt; pStmt = pStmt->tail )
  {
    if( pStmt->head )
      semanticCheckStmt( pStmt->head, pScope );
  } /* for */
}

void semanticCheck( Decl *prog )
{
  DeclList *pScope = NULL;

  pScope = (DeclList*) malloc( sizeof( DeclList ) );
  pScope->head = prog;
  pScope->tail = NULL;

  semanticCheckDeclList( prog->pDecls, pScope );
  semanticCheckStmtList( prog->pStmts, pScope );
}